# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for bluetooth service controls capability."""
import abc

from gazoo_device.capabilities.interfaces import capability_base


class BluetoothServiceBase(capability_base.CapabilityBase):
  """Abstract base class for bluetooth service controls capability."""

  @abc.abstractmethod
  def status(self) -> bool:
    """Returns if the bluetooth service is active or not."""

  @abc.abstractmethod
  def start(self) -> None:
    """Starts the bluetooth service."""

  @abc.abstractmethod
  def stop(self) -> None:
    """Stops the bluetooth service."""

  @abc.abstractmethod
  def restart(self) -> None:
    """Restarts the bluetooth service."""
